# Copyright (C) 2004, 2005 Gregoire Lejeune <gregoire.lejeune@free.fr>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

# This lib is a fork of the HTMLSSI lib from Domingo Alvarez Duarte
# <domingo@dad-it.com>

require 'net/http'
require 'uri'

# <!--#config sizefmt="value" timefmt="value" -->
SSI_config = 'config'

# <!--#include virtual|file="file" -->
SSI_include = 'include'
# <!--#exec cmd|cgi="file" --> ## PARTIAL SUPPORT
SSI_exec = 'exec'

# <!--#fsize virtual|file="file" --> ## TODO
SSI_fsize = 'fsize'
# <!--#flastmod virtual|file="file" --> ## TODO
SSI_flastmod = 'flastmod'

# <!--#break --> ## TODO
SSI_break = 'break'
# <!--#goto="name" --> ## TODO
SSI_goto = 'goto'
# <!--#label="name" --> ## TODO
SSI_label = 'label'

# <!--#if expr="cond" -->
SSI_if = 'if'
# <!--#elif expr="cond" -->
SSI_elif = /elif|elsif|elseif/
# <!--#else -->
SSI_else = 'else'
# <!--#endif -->
SSI_endif = 'endif'

# <!--#set var="name" value="value" --> 
SSI_set = 'set'
# <!--#echo var="name" -->
SSI_echo = 'echo'

# <!--#to_var var="name" -->
SSI_to_var = 'to_var'
SSI_endto_var = 'endto_var'

# <!--#line_vars -->
SSI_line_vars = 'line_vars'
SSI_endline_vars = 'endline_vars'

# <!--#count -->
SSI_count = 'count'

# <!--#loop expr="cond"-->
SSI_loop = 'loop'
SSI_endloop = 'endloop'

# <!--#comment -->
SSI_comment = 'comment'
SSI_endcomment = 'endcomment'

# <!--#cpt name="name" init="value"|incr="value" --> ## TODO
SSI_cpt = 'cpt'

# deafult tag shortcut for variables
SSI_search_re = /<!--#(\w+)(.*?)-->|\[\)\s*(\w+)\s*\(\]/

module HTML
  class SSI
  
    attr_writer :vars,:fget_var,:fget_include
    
    def initialize( bStrip = true )
      @vars = {}
      @cpts = {}
      @fget_var = nil
      @fget_include = nil
      @count = 0
      @bStrip = bStrip
      @bBreak = false
      
      # Config
      @hxConfig = Hash::new( )
      @hxConfig["sizefmt"] = "abbrev"
      @hxConfig["timefmt"] = "%a %b %d %H:%M:%S %Z %Y"
      
      @hxSizeFmtTable = Hash::new( )
      # -- sizefmt --
      @hxSizeFmtTable["abbrev"] = { "unit" => "k", "mul" => 1 }
      @hxSizeFmtTable["bytes"] = { "unit" => "", "mul" => 1024 }
    end
    
    private
    
    def nonull( xData )
      if @bStrip == true 
        if xData.strip.length > 0
          return( xData )
        end
        return ( '' )
        end
      return( xData )
    end
  
    def getConfig( xConf, xValue )
      case xConf 
        when "sizefmt"
          if @hxSizeFmtTable.keys.include?( xValue ) == true
            @hxConfig["sizefmt"] = xValue
          end
          return
          
        when "timefmt"
          @hxConfig["timefmt"] = xValue
          return
          
        else
          return
      end
      
      return
    end
    
    def getExec( xType, xCmd )
      if xType == "cmd"
        return( `#{xCmd}` )
      elsif xType == "cgi"
        return( "<!-- exec cgi='#{xCmd}' : not yet supported -->" )
      end
      
      return( "<!-- include #{xType}='#{xCmd}' : type #{xType} unknown -->" )
    end
    
    def getInclude( xType, xFileName )
      xFileContent = "<!-- include ERROR -->\n"
      if @fget_include
        xFileContent = @fget_include.call( xType, xFileName )
      else
        if xType == "file"
          hFile = File.open( xFileName.untaint )
          xFileContent = hFile.read
          hFile.close
        elsif xType == "virtual"
          xFileContent = Net::HTTP.get( URI.parse( xFileName.untaint ).untaint ).untaint
        else
          xFileContent = "<!-- include #{xType}='#{xFileName}' : type #{xType} unknown -->"
        end
      end
      xFileContent
    end
    
    def getFsize( xType, xFileName )
      return( ( File::size( xFileName ) * @hxSizeFmtTable[@hxConfig["sizefmt"]]["mul"] ).to_s + @hxSizeFmtTable[@hxConfig["sizefmt"]]["unit"] )
    end
    
    def getFlastmod( xType, xFileName )
      return( ( File::ctime( xFileName ).strftime( @hxConfig["timefmt"] ).to_s ) )
    end
    
    def getVar( xVarName )
      xVarContent = @vars[xVarName]
      if !xVarContent
        if @fget_var
          xVarContent = @fget_var.call( xVarName )
        elsif ENV[xVarName]
          xVarContent = ENV[xVarName]
        else
          return "<!--|#{xVarName}|-->"
        end
      end
      return( xVarContent )
    end
    
    def getBlock( xHTMLData, iStartPosition, xStartTag, xEndTag, xAlternateEndTag = '' )
      iNestedTag = 0
      bTagFound = false
      iCurrentPosition = iStartPosition
      lvResult = []
      
      while( iCurrentPosition ) do
        iCurrentPosition = xHTMLData.index(SSI_search_re, iCurrentPosition)
        xCommand = $1 ? $1.downcase : $3.downcase
        xTag = $&
        case xCommand
          when xStartTag
            iNestedTag += 1
          when xAlternateEndTag
            if iNestedTag == 0
              bTagFound = true
              break 
            end
          when xEndTag
            iNestedTag -= 1            
            if iNestedTag < 0
              bTagFound = true
              break 
            end
        end        
        iCurrentPosition += xTag.size
      end
      if bTagFound
        lvResult[0] = iCurrentPosition
        lvResult[1] = xTag.size
        lvResult[2] = xCommand
        lvResult[3] = xHTMLData[iStartPosition..iCurrentPosition-1]
      end
      lvResult
    end
    
    def expressionResult( xExpression )
      mExprMatching = xExpression.scan( /(["']?)(\$?\S+)\1\s*(==|!=|>|<|<=|>=)\s*(["']?)(\$?\S+)\4/ )[0]
      #print "ss-#{xExpression}-ss"
      #print mExprMatching.join('::')

      vValueOne = mExprMatching[0].size > 0  ? mExprMatching[1] : mExprMatching[1][0] == '$'[0] ? getVar(mExprMatching[1][1..-1]) : mExprMatching[1]
      vValueTwo = mExprMatching[3].size > 0  ? mExprMatching[4] : mExprMatching[4][0] == '$'[0] ? getVar(mExprMatching[4][1..-1]) : mExprMatching[4]

      if vValueTwo.class != vValueOne.class
        vValueOne = vValueOne.to_f
        vValueTwo = vValueTwo.to_f
      end
      
      case mExprMatching[2]
        when '=='
          vValueOne == vValueTwo
        when '!='
          vValueOne != vValueTwo
        when '<'
          vValueOne < vValueTwo
        when '>'    
          vValueOne > vValueTwo
        when '<='
          vValueOne <= vValueTwo
        when '>='
          vValueOne >= vValueTwo
        else
          false
      end
    end    
  
    public
    
    def parse( xHTMLData )
      xParseResult = ''
      iLastPosition = 0
      iCurrentPosition = 0
      
      while( iCurrentPosition ) do
        iCurrentPosition = xHTMLData.index( SSI_search_re, iLastPosition )
        
        if iCurrentPosition
          xCommand = $1 ? $1.downcase : $3.downcase # command or var name
          xCommandParameters = $2         # command parameters
          iTagSize = $&.size        # match tag size
          
          if iCurrentPosition > 0
            xParseResult << nonull( xHTMLData[iLastPosition..iCurrentPosition-1] ) # anithing else is a literal
          end
          
          case xCommand
            when SSI_config
              xCommandParameters.gsub( /(\w+)\s*=\s*(["'])(.*?)\2/ ) { 
                getConfig( $1, $3 )
              }
              
            when SSI_include
              xParseResult << nonull( xCommandParameters.gsub( /(\w+)\s*=\s*(["'])(.*?)\2/ ) { 
                parse( getInclude( $1, $3 ) )
              } )
  
            when SSI_exec
              xParseResult << nonull( xCommandParameters.gsub( /(\w+)\s*=\s*(["'])(.*?)\2/ ) { 
                parse( getExec( $1, $3 ) )
              } )

            when SSI_fsize
              xParseResult << nonull( xCommandParameters.gsub( /(\w+)\s*=\s*(["'])(.*?)\2/ ) { 
                parse( getFsize( $1, $3 ) )
              } )
              
            when SSI_flastmod
              xParseResult << nonull( xCommandParameters.gsub( /(\w+)\s*=\s*(["'])(.*?)\2/ ) { 
                parse( getFlastmod( $1, $3 ) )
              } )
  
            when SSI_break
              @bBreak = true
              return( xParseResult )
              
            when SSI_goto
              xParseResult << "<!-- SSI goto not supported yet ! sorry -->"
  
            when SSI_label
              xParseResult << "<!-- SSI label not supported yet ! sorry -->"
  
            when SSI_count
              xParseResult << @count.to_s
              @count = @count + 1
  
            when SSI_if, SSI_elif
              xExpression = xCommandParameters.sub( /expr\s*=\s*(["'])(.*?)\1/i, '\2' ) 
              lvIFBlock = getBlock( xHTMLData, iCurrentPosition+iTagSize, SSI_if, SSI_endif, /#{SSI_elif}|#{SSI_else}/ )

              from = lvIFBlock[1]
              if expressionResult( xExpression ) 
                xParseResult << nonull( parse( lvIFBlock[3] ) )
                if lvIFBlock[2] == SSI_endif
                  iCurrentPosition = lvIFBlock[0]
                  iTagSize = lvIFBlock[1]
                else    # consume elif,else blocks
                  lvELSEBlock = getBlock( xHTMLData, lvIFBlock[0]+lvIFBlock[1], SSI_if, SSI_endif )
                  iCurrentPosition = lvELSEBlock[0]
                  iTagSize = lvELSEBlock[1]
                end
              else
                iCurrentPosition = lvIFBlock[0]
                iTagSize = (lvIFBlock[2] == SSI_endif) ? iTagSize = lvIFBlock[1] : 0
                # next tag can be elif,else
              end
              
            when SSI_else
              lvELSEBlock = getBlock( xHTMLData, iCurrentPosition+iTagSize, SSI_if, SSI_endif )
              xParseResult << nonull( parse( lvELSEBlock[3] ) )
              iCurrentPosition = lvELSEBlock[0]
              iTagSize = lvELSEBlock[1]
            
            when SSI_loop
              xExpression = xCommandParameters.sub( /expr\s*=\s*(["'])(.*?)\1/i, '\2' )

              lvLOOPBlock = getBlock( xHTMLData, iCurrentPosition+iTagSize, xCommand, SSI_endloop )
              xLOOPBody = lvLOOPBlock[3]
              while expressionResult( xExpression ) do
                xParseResult << nonull( parse( xLOOPBody ) )
                if @bBreak == true
                  @bBreak = false
                  break
                end
              end
              iCurrentPosition = lvLOOPBlock[0]
              iTagSize = lvLOOPBlock[1]
            
            when SSI_to_var # save a block of text to a var
              xVar = xCommandParameters.sub( /var\s*=\s*(["'])(\w+)\1/i, '\2' ).strip
              lvVARBlock = getBlock( xHTMLData, iCurrentPosition+iTagSize, '', SSI_endto_var )
              @vars[xVar] = lvVARBlock[3]
              iCurrentPosition = lvVARBlock[0]
              iTagSize = lvVARBlock[1]
            
            when SSI_line_vars # save a block vars (one by line)
              lvVARBlock = getBlock( xHTMLData, iCurrentPosition+iTagSize, '', SSI_endline_vars )
              lvVARBlock[3].scan( /(\w+)\s*=(.*?)$/ ).each {|ar| @vars[ar[0]]=ar[1]}
              iCurrentPosition = lvVARBlock[0]
              iTagSize = lvVARBlock[1]
            
            when SSI_comment # cut a section of a page
              lvVARBlock = getBlock( xHTMLData, iCurrentPosition+iTagSize, '', SSI_endcomment )
              iCurrentPosition = lvVARBlock[0]
              iTagSize = lvVARBlock[1]
            
            when SSI_set
              xCommandParameters.gsub( /var\s*=\s*(["'])(.*?)\1\s*value\s*=\s*(["'])(.*?)\3/ ) { @vars[$2] = $4 }
              #xCommandParameters.gsub( /(\w+)\s*=\s*(["'])(.*?)\2/ ) { @vars[$1] = $3 }
              #xParseResult << nonull( xCommandParameters.gsub( /(\w+)\s*=\s*(["'])(.*?)\2/ ) { @vars[$1] = $3 } )
            
            when SSI_echo
              xParseResult << nonull( xCommandParameters.gsub( /var\s*=\s*(["'])(\w+)\1/i ) { getVar( $2 ) } )
            
            when SSI_cpt
              mCptName = xCommandParameters.scan( /name\s*=\s*(["'])(\w+)\1/i )[0]
              mCptInit = xCommandParameters.scan( /init\s*=\s*(["'])(\w+)\1/i )[0]
              mCptIncr = xCommandParameters.scan( /incr\s*=\s*(["'])(\w+)\1/i )[0]
              if mCptName
                if mCptInit
                  @cpts[mCptName[1]] = mCptInit[1].to_i
                end
              
                if @cpts[mCptName[1]] and mCptIncr
                  @cpts[mCptName[1]] = @cpts[mCptName[1]] + mCptIncr[1].to_i
                end
                
                @vars[mCptName[1]] = @cpts[mCptName[1]]
              end
              
            else
              if $3    # we assume this is some var betwen [)some_var(]
                xParseResult << nonull( getVar( $3 ) )
              end
          end
          iLastPosition = iCurrentPosition + iTagSize
        end
      end
      
      if iLastPosition < xHTMLData.size
        xParseResult << nonull( xHTMLData[iLastPosition..-1] ) # remainder literal
      end
      
      xParseResult
    end
  end
end
